# Copyright 2019 Kensho Technologies, LLC.
"""Module for Wikidata linked data interface endpoints."""
import logging

import requests
from qwikidata import typedefs

logger = logging.getLogger(__name__)
WIKIDATA_LDI_URL = "https://www.wikidata.org/wiki/Special:EntityData"
VALID_ENTITY_PREFIXES = ("Q", "P", "L")


class LdiResponseNotOk(Exception):
    pass


class InvalidEntityId(Exception):
    pass


def get_entity_dict_from_api(
    entity_id: typedefs.EntityId, base_url: str = WIKIDATA_LDI_URL
) -> typedefs.EntityDict:
    """Get a dictionary representing a wikidata entity from the linked data interface API.

    https://www.wikidata.org/wiki/Wikidata:Data_access#Linked_Data_interface

    Parameters
    ----------
    entity_id
      A Wikidata entity id beginning with "Q", "P", or "L" (e.g. "Q42")
    base_url
      The linked data interface URL to use

    Examples
    --------
    Get the entity dictionary for item Q42,

    ::

      >>> entity_dict = get_entity_dict_from_api('Q42')
      >>> pprint(entity_dict, indent=4, depth=1)
      {   'aliases': {...},
          'claims': {...},
          'descriptions': {...},
          'id': 'Q42',
          'labels': {...},
          'lastrevid': 716282445,
          'modified': '2018-07-27T08:03:25Z',
          'ns': 0,
          'pageid': 138,
          'sitelinks': {...},
          'title': 'Q42',
          'type': 'item'}}}

    """
    if not isinstance(entity_id, str):
        raise InvalidEntityId(
            'entity_id must be a string (e.g. "Q42") but got entity_id={}.'.format(entity_id)
        )
    if not entity_id[0] in VALID_ENTITY_PREFIXES:
        raise InvalidEntityId(
            "entity_id must start with one of {} but got entity_id={}.".format(
                VALID_ENTITY_PREFIXES, entity_id
            )
        )

    url = "{}/{}.json".format(base_url, entity_id)
    response = requests.get(url)
    if response.ok:
        entity_dict_full = response.json()
    else:
        raise LdiResponseNotOk(
            "input entity id: {}, "
            "response.headers: {}, "
            "response.status_code: {}, "
            "response.text: {}".format(
                entity_id, response.headers, response.status_code, response.text
            )
        )

    # remove redundant top level keys
    returned_entity_id = next(iter(entity_dict_full["entities"]))
    entity_dict = entity_dict_full["entities"][returned_entity_id]

    if entity_id != returned_entity_id:
        logger.warning(
            "Wikidata redirect detected.  Input entity id={}. Returned entity id={}.".format(
                entity_id, returned_entity_id
            )
        )

    return entity_dict



# Copyright 2019 Kensho Technologies, LLC.
"""Module for the Wikidata SPARQL endpint."""
from typing import Dict, List, Union

import requests

WIKIDATA_SPARQL_URL = "https://query.wikidata.org/sparql"


def return_sparql_query_results(
    query_string: str, wikidata_sparql_url: str = WIKIDATA_SPARQL_URL
) -> Dict:
    """Send a SPARQL query and return the JSON formatted result.

    Parameters
    ----------
    query_string: str
      SPARQL query string
    wikidata_sparql_url: str, optional
      wikidata SPARQL endpoint to use
    """
    return requests.get(
        wikidata_sparql_url, params={"query": query_string, "format": "json"}
    ).json()


def get_subclasses_of_item(item_id: str, return_qids: bool = True) -> Union[List[str], Dict]:
    """Return all subclasses of a wikidata item.

    Finds all items where a chain of the following form exists,::

      Qid_i -[P279]-> Qid_j ... Qid_k -[P279]-> item_id

    Will always include the item itself in the return results.  Note that
    property P279 = "subclass of".

    Parameters
    ----------
    item_id: str
      The item to use as the end of the chain.
    return_qids: bool, optional
      If false, the SPARQL query result is returned unaltered.  If true,
      a list of item id string is returned instead.  See examples.

    Examples
    --------
    We can get all item IDs that are subclasses of `Q6256`,

    ::

      >>> get_subclasses_of_item('Q6256')
      ['Q6256',
       'Q112099',
       'Q123480',
       ...
       'Q4994005',
       'Q6805624',
       'Q15895923']

    ::

      >>> get_subclasses_of_item('Q6256', return_qids=False)
      {'head': {'vars': ['WDid']},
       'results': {'bindings': [
         {'WDid': {'type': 'uri',
           'value': 'http://www.wikidata.org/entity/Q6256'}},
         {'WDid': {'type': 'uri',
           'value': 'http://www.wikidata.org/entity/Q112099'}},
         ...
         {'WDid': {'type': 'uri',
           'value': 'http://www.wikidata.org/entity/Q6805624'}},
         {'WDid': {'type': 'uri',
           'value': 'http://www.wikidata.org/entity/Q15895923'}}]}}
    """
    query_string = """
    SELECT $WDid
    WHERE {{
      ?WDid (wdt:P279)* wd:{} .
    }}
    """.format(
        item_id
    )
    results = return_sparql_query_results(query_string)
    if return_qids:
        uris = [binding["WDid"]["value"] for binding in results["results"]["bindings"]]
        qids = [uri.split("/")[-1] for uri in uris]
        return qids
    else:
        return results



# Copyright 2019 Kensho Technologies, LLC.
"""Module for Wikidata JSON dumps."""
import bz2
import gzip
import json
import logging
import os
import subprocess
from contextlib import contextmanager
from typing import IO, Any, Dict, Iterator, List, Optional, Tuple


class WikidataJsonDump:
    """Class for Wikidata JSON dump files.

    Represents a json file from https://dumps.wikimedia.org/wikidatawiki/entities.
    File names are of the form "wikidata-YYYYMMDD-all.json[.bz2|.gz]".  The file is a single JSON
    array and there is one element (i.e. item or property) on each line with the first and
    last lines being the opening and closing square brackets.  This class can handle bz2 or gz
    compressed files as well as the uncompressed json files.

    Parameters
    ----------
    filename: str
      The wikidata JSON dump file name (e.g. `my_data_dir/wikidata-20180730-all.json.bz2`)
    """

    def __init__(self, filename: str) -> None:
        if not isinstance(filename, str):
            raise ValueError("filename must be a string")

        if filename.endswith(".json"):
            self.basename, _ = os.path.splitext(filename)
            self.compression = None
        elif filename.endswith(".json.bz2"):
            self.basename, _ = os.path.splitext(os.path.splitext(filename)[0])
            self.compression = "bz2"
        elif filename.endswith(".json.gz"):
            self.basename, _ = os.path.splitext(os.path.splitext(filename)[0])
            self.compression = "gz"
        else:
            raise ValueError('filename must end with ".json.bz2" or ".json.gz" or ".json"')

        self.filename = filename
        self.logger = logging.getLogger(__name__)

    @contextmanager
    def _open_dump_file(self) -> Iterator[IO[Any]]:
        """Context manager that opens compressed/uncompressed dump files.

        It is important to open the file in binary mode even if it is not compressed. This allows us
        to handle decoding in one place.
        """
        if self.compression == "bz2":
            with bz2.open(self.filename, mode="rb") as fp:
                yield fp
        elif self.compression == "gz":
            with gzip.open(self.filename, mode="rb") as fp:
                yield fp
        else:
            with open(self.filename, mode="rb") as fp:
                yield fp

    def iter_lines(self) -> Iterator[str]:
        """Generate lines from JSON dump file."""
        with self._open_dump_file() as fp:
            for linebytes in fp:
                yield linebytes.decode("utf-8")

    def __iter__(self) -> Iterator[Dict]:
        """Iterate over lines in the file."""
        for line_str in self.iter_lines():
            line_str = line_str.rstrip(",\n")
            # first and last lines are opening and closing brackets
            if line_str in ["[", "]"]:
                continue
            yield json.loads(line_str)

    def _write_chunk(
        self, out_fbase: str, ichunk: int, out_lines: List[str]
    ) -> Tuple[List[str], int, str]:
        """Write a single chunk to disk."""
        out_fname = "{}-ichunk_{}.json".format(out_fbase, ichunk)
        self.logger.debug("writing {}".format(out_fname))
        out_lines = [out_line.rstrip(",\n") for out_line in out_lines]
        with open(out_fname, "w") as fp:
            fp.write("[\n")
            fp.write(",\n".join(out_lines))
            fp.write("\n]\n")

        if self.compression == "bz2":
            args = ["bzip2", out_fname]
            subprocess.check_output(args)
            out_fname = "{}.bz2".format(out_fname)
        elif self.compression == "gz":
            args = ["gzip", out_fname]
            subprocess.check_output(args)
            out_fname = "{}.gz".format(out_fname)

        out_lines = []
        ichunk += 1
        return out_lines, ichunk, out_fname

    def create_chunks(
        self,
        out_fbase: Optional[str] = None,
        num_lines_per_chunk: int = 100,
        max_chunks: int = 10 ** 10,
    ) -> List[str]:
        """Produce N files with `num_lines_per_chunk` wikidata items per file.

        Parameters
        ----------
        out_fbase: str
          Each output file will have the form `{out_fbase}_ichunk_{ichunk}.json[.bz2|.gz]`
        num_lines_per_chunk: int
          Number of lines per chunk file
        max_chunks: int
          Maximum number of chunks to write
        """
        wd_dump = WikidataJsonDump(self.filename)
        if out_fbase is None:
            out_fbase = self.basename

        ichunk = 0
        out_lines = []  # type: List[str]
        out_fnames = []  # type: List[str]

        for iline, line in enumerate(wd_dump.iter_lines()):
            if line.strip() in ["[", "]"]:
                continue
            out_lines.append(line)
            if len(out_lines) >= num_lines_per_chunk:
                out_lines, ichunk, out_fname = self._write_chunk(out_fbase, ichunk, out_lines)
                out_fnames.append(out_fname)

            if ichunk >= max_chunks:
                return out_fnames

        if len(out_lines) > 0:
            out_lines, ichunk, out_fname = self._write_chunk(out_fbase, ichunk, out_lines)
            out_fnames.append(out_fname)

        return out_fnames

    def __str__(self) -> str:
        return "WikidataJsonDump(filename={})".format(self.filename)

    def __repr__(self) -> str:
        return self.__str__()



# Copyright 2019 Kensho Technologies, LLC.
"""qwikidata utilities."""

import itertools
import json
from typing import Iterable, Iterator, Tuple

from qwikidata.entity import WikidataEntity


def pairwise(iterable: Iterable) -> Iterator[Tuple]:
    """Return pairwise tuples s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def dump_entities_to_json(entities: Iterable[WikidataEntity], out_fname: str) -> None:
    """Write entities to JSON file.

    Parameters
    ----------
    entities
      An iterable of instances of WikidataEntity
    out_fname
      Output file name
    """
    with open(out_fname, "w") as fp:
        fp.write("[")
        ent_iter = iter(entities)
        ent = next(ent_iter, None)
        while ent:
            ent_str = json.dumps(ent._entity_dict)
            fp.write("\n{}".format(ent_str))
            ent = next(ent_iter, None)
            if ent: fp.write(",")

        fp.write("\n]")



# Copyright 2019 Kensho Technologies, LLC.
"""Module for Wikidata Snaks."""

from typing import Union

from qwikidata import typedefs
from qwikidata.datavalue import WikidataDatavalue, get_datavalue_from_snak_dict


class WikidataSnak:
    """A Wikidata snak.

    Parameters
    ----------
    snak_dict: dict
      A dictionary representing a Wikidata snak.
      See `the wikibase JSON data model docs`_ for a description
      of the format.


    Attributes
    ----------
    snaktype: str
      One of ["value", "somevalue", "novalue"].  "value" indicates that a known value exists,
      "somevalue" indicates that an unknown value exists, and "novalue" indicates that no
      value exists.
    property_id: str
      A Wikidata property id (e.g. "P551")
    snak_datatype: str or `None`
      The snak data type (SDT). Must be one of ["commonsMedia", "external-id", "geo-shape",
      "globe-coordinate", "math", "monolingualtext", "quantity", "string", "tabular-data", "time",
      "url", "wikibase-item", "wikibase-property", `None`].  Will be `None` if `snaktype` is not
      "value".
    value_datatype: str or `None`
      The value data type (VDT).  Must be one of ["globecoordinate", "monolingualtext",
      "quantity", "string", "time", "wikibase-entityid", `None`].  Will be `None` if `snaktype`
      is not "value".
    datavalue: :py:class:`WikidataDataValue` or `None`
      The datavalue object for this snak.  Has `datatype` and `value` attributes. (this data type
      is the same as VDT above).


    .. note::

      There are two related data types here, the snak data type (SDT) and the value data
      type (VDT).  The SDT is derived from the data type of the property referenced by
      `property_id` and is stored in the `snak_datatype` attribute.  The VDT defines
      the structure of the `value` attribute of :py:class:`WikidataDataValue`.
      The VDT does not allow for interpretation of the datavalue, only for processing of the raw
      structure. As an example, a link to a web page may have SDT="url", but have VDT="string".


    .. _the wikibase JSON data model docs: https://www.mediawiki.org/wiki/Wikibase/DataModel/JSON
    """

    def __init__(self, snak_dict: typedefs.SnakDict) -> None:
        self._validate_snak_dict(snak_dict)
        self._snak_dict = snak_dict

    def _validate_snak_dict(self, snak_dict: typedefs.SnakDict) -> None:
        """Raise excpetions if snak_dict is not valid."""
        _REQUIRED_KEYS = ["snaktype", "property"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in snak_dict:
                raise ValueError(
                    "required snak_dict keys are {} but only found {}".format(
                        _REQUIRED_KEYS, list(snak_dict.keys())
                    )
                )
        self.snaktype = snak_dict["snaktype"]
        self.property_id = snak_dict["property"]

        self.snak_datatype = None  # type: Union[str, None]
        self.value_datatype = None  # type: Union[str, None]
        self.datavalue = None  # type: Union[WikidataDatavalue, None]

        if self.snaktype == "value":
            _REQUIRED_KEYS = ["datavalue", "datatype"]
            for req_key in _REQUIRED_KEYS:
                if req_key not in snak_dict:
                    raise ValueError(
                        "required snak_dict keys are {} but only found {}".format(
                            _REQUIRED_KEYS, list(snak_dict.keys())
                        )
                    )
            self.snak_datatype = snak_dict["datatype"]
            self.value_datatype = str(snak_dict["datavalue"]["type"])
            self.datavalue = get_datavalue_from_snak_dict(snak_dict)

        elif self.snaktype == "somevalue" or self.snaktype == "novalue":
            self.snak_datatype = None
            self.value_datatype = None
            self.datavalue = None

        else:
            raise ValueError(
                'snaktype must be one of ["value", "somevalue", "novalue"] but got {}'.format(
                    self.snaktype
                )
            )

    def __str__(self) -> str:
        return (
            "WikidataSnak(snaktype={}, property_id={}, "
            "snak_datatype={}, value_datatype={}, "
            "datavalue={})".format(
                self.snaktype,
                self.property_id,
                self.snak_datatype,
                self.value_datatype,
                self.datavalue,
            )
        )

    def __repr__(self) -> str:
        return self.__str__()



# Copyright 2019 Kensho Technologies, LLC.
"""Module providing Wikidata Types."""
from typing import Dict, List, NewType, Union

from mypy_extensions import TypedDict

# Ids
# ====================================================
ItemId = NewType("ItemId", str)
PropertyId = NewType("PropertyId", str)
LexemeId = NewType("LexemeId", str)
EntityId = Union[ItemId, PropertyId, LexemeId]

NumericItemId = NewType("NumericItemId", int)
NumericPropertyId = NewType("NumericPropertyId", int)
NumericLexemeId = NewType("NumericLexemeId", int)
NumericEntityId = Union[NumericItemId, NumericPropertyId, NumericLexemeId]

FormId = NewType("FormId", str)
SenseId = NewType("SenseId", str)


# Named strings
# ====================================================
EntityUri = NewType("EntityUri", str)
LanguageCode = NewType("LanguageCode", str)


# Datavalues
# ====================================================
GlobeCoordinateValue = TypedDict(
    "GlobeCoordinateValue",
    {"latitude": float, "longitude": float, "precision": float, "globe": EntityUri},
)
GlobeCoordinateDatavalueDict = TypedDict(
    "GlobeCoordinateDatavalueDict", {"type": str, "value": GlobeCoordinateValue}
)

MonolingualTextValue = TypedDict("MonolingualTextValue", {"text": str, "language": LanguageCode})
MonolingualTextDatavalueDict = TypedDict(
    "MonolingualTextDatavalueDict", {"value": MonolingualTextValue, "type": str}
)

QuantityValue = TypedDict(
    "QuantityValue", {"amount": str, "upperBound": str, "lowerBound": str, "unit": str}
)
QuantityDatavalueDict = TypedDict("QuantityDatavalueDict", {"value": QuantityValue, "type": str})

StringValue = NewType("StringValue", str)
StringDatavalueDict = TypedDict("StringDatavalueDict", {"value": StringValue, "type": str})

TimeValue = TypedDict(
    "TimeValue",
    {
        "time": str,
        "timezone": int,
        "calendarmodel": EntityUri,
        "precision": int,
        "before": int,
        "after": int,
    },
)
TimeDatavalueDict = TypedDict("TimeDatavalueDict", {"value": TimeValue, "type": str})

WikibaseEntityIdValue = TypedDict(
    "WikibaseEntityIdValue", {"entity-type": str, "id": EntityId, "numeric-id": NumericEntityId}
)
WikibaseEntityIdDatavalueDict = TypedDict(
    "WikibaseEntityIdDatavalueDict", {"value": WikibaseEntityIdValue, "type": str}
)

WikibaseUnmappedEntityIdValue = NewType("WikibaseUnmappedEntityIdValue", str)
WikibaseUnmappedEntityIdDatavalueDict = TypedDict(
    "WikibaseUnmappedEntityIdDatavalueDict", {"value": WikibaseUnmappedEntityIdValue, "type": str}
)


DatavalueDict = Union[
    GlobeCoordinateDatavalueDict,
    MonolingualTextDatavalueDict,
    QuantityDatavalueDict,
    StringDatavalueDict,
    TimeDatavalueDict,
    WikibaseEntityIdDatavalueDict,
    WikibaseUnmappedEntityIdDatavalueDict,
]


# Snak
# ====================================================
SnakDict = TypedDict(
    "SnakDict",
    {"snaktype": str, "property": PropertyId, "datatype": str, "datavalue": DatavalueDict},
)

# Claims
# ====================================================
ReferenceDict = TypedDict(
    "ReferenceDict",
    {"hash": str, "snaks": Dict[PropertyId, List[SnakDict]], "snaks-order": List[PropertyId]},
)

QualifierDict = TypedDict(
    "QualifierDict",
    {
        "hash": str,
        "snaktype": str,
        "property": PropertyId,
        "datatype": str,
        "datavalue": DatavalueDict,
    },
)

ClaimDict = TypedDict(
    "ClaimDict",
    {
        "id": str,
        "mainsnak": SnakDict,
        "type": str,
        "rank": str,
        "qualifiers": Dict[PropertyId, List[QualifierDict]],
        "references": List[ReferenceDict],
        "qualifiers-order": List[PropertyId],
    },
)

ClaimList = List[ClaimDict]


# Items and Properties
# ====================================================
LabelDict = TypedDict("LabelDict", {"language": LanguageCode, "value": str})

DescriptionDict = TypedDict("DescriptionDict", {"language": LanguageCode, "value": str})

AliasDict = TypedDict("AliasDict", {"language": LanguageCode, "value": str})

AliasList = List[AliasDict]

SitelinkDict = TypedDict(
    "SitelinkDict", {"site": str, "title": str, "badges": List[str], "url": str}
)

ItemDict = TypedDict(
    "ItemDict",
    {
        "pageid": int,
        "ns": int,
        "title": str,
        "lastrevid": int,
        "modified": str,
        "type": str,
        "id": ItemId,
        "labels": Dict[LanguageCode, LabelDict],
        "descriptions": Dict[LanguageCode, DescriptionDict],
        "aliases": Dict[LanguageCode, AliasList],
        "sitelinks": Dict[str, SitelinkDict],
        "claims": Dict[PropertyId, ClaimList],
    },
)

PropertyDict = TypedDict(
    "PropertyDict",
    {
        "pageid": int,
        "ns": int,
        "title": str,
        "lastrevid": int,
        "modified": str,
        "type": str,
        "id": PropertyId,
        "labels": Dict[LanguageCode, LabelDict],
        "descriptions": Dict[LanguageCode, DescriptionDict],
        "aliases": Dict[LanguageCode, AliasList],
        "claims": Dict[PropertyId, ClaimList],
    },
)


# Lexemes
# ====================================================
RepresentationDict = TypedDict("RepresentationDict", {"language": LanguageCode, "value": str})

FormDict = TypedDict(
    "FormDict",
    {
        "id": FormId,
        "representations": Dict[LanguageCode, RepresentationDict],
        "grammaticalFeatures": List[ItemId],
        "claims": Dict[PropertyId, ClaimList],
    },
)


GlossDict = TypedDict("GlossDict", {"language": LanguageCode, "value": str})

SenseDict = TypedDict(
    "SenseDict",
    {
        "id": SenseId,
        "glosses": Dict[LanguageCode, GlossDict],
        "claims": Dict[PropertyId, ClaimList],
    },
)

LemmaDict = TypedDict("LemmaDict", {"language": LanguageCode, "value": str})

LexemeDict = TypedDict(
    "LexemeDict",
    {
        "pageid": int,
        "ns": int,
        "title": str,
        "lastrevid": int,
        "modified": str,
        "type": str,
        "id": LexemeId,
        "lemmas": Dict[LanguageCode, LemmaDict],
        "lexicalCategory": ItemId,
        "language": ItemId,
        "claims": Dict[PropertyId, ClaimList],
        "forms": List[FormDict],
        "senses": List[SenseDict],
    },
)

EntityDict = Union[ItemDict, PropertyDict, LexemeDict]



import codecs
import os
import re

from setuptools import find_packages, setup

#  https://packaging.python.org/guides/single-sourcing-package-version/
#  #single-sourcing-the-version


def read_file(filename: str) -> str:
    """Read package file as text to get name and version"""
    # intentionally *not* adding an encoding option to open
    # see here:
    # https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, "qwikidata", filename), "r") as f:
        return f.read()


def find_version() -> str:
    """Only define version in one place"""
    version_file = read_file("__init__.py")
    version_match = re.search(r'^__version__ = ["\']([^"\']*)["\']', version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def find_name() -> str:
    """Only define name in one place"""
    name_file = read_file("__init__.py")
    name_match = re.search(r'^__package_name__ = ["\']([^"\']*)["\']', name_file, re.M)
    if name_match:
        return name_match.group(1)
    raise RuntimeError("Unable to find name string.")


def find_long_description() -> str:
    """Return the content of the README.rst file."""
    return read_file("../README.rst")


setup(
    name=find_name(),
    version=find_version(),
    description="Manipulate Wikidata Objects.",
    long_description=find_long_description(),
    long_description_content_type="text/x-rst",
    url="https://github.com/kensho-technologies/qwikidata",
    author="Kensho Technologies LLC.",
    author_email="qwikidata-maintainer@kensho.com",
    license="Apache 2.0",
    packages=find_packages(exclude=["tests*"]),
    install_requires=["mypy-extensions", "requests"],
    extras_require={
        "dev": [
            "pre-commit",
            "pytest",
            "pytest-cov",
            "sphinx",
            "sphinx_rtd_theme",
            "sphinx_autodoc_typehints",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="wikidata parser open data",
    python_requires=">=3.5",
)



# Copyright 2019 Kensho Technologies, LLC.
"""Module for Wikidata Claims (aka Statements)."""

from collections import OrderedDict
from collections.abc import Sequence
from typing import List, Union, overload

from qwikidata import typedefs
from qwikidata.snak import WikidataSnak


class WikidataReference:
    """A reference about a claim about a Wikidata Entity.

    See: https://www.wikidata.org/wiki/Help:Sources

    This class can be initialized from an entity dictionary as,

    .. code-block:: python

      >>> reference_dict = q42_dict['claims']['P69'][0]['references'][0]
      >>> wikidata_reference = WikidataReference(reference_dict)


    Parameters
    ----------
    reference_dict
      A dictionary representing a Wikidata reference.
      See `the wikibase JSON data model docs`_ for a description
      of the format.


    Attributes
    ----------
    referencehash: str
      Unique id for this reference
    snaks: collections.OrderedDict
      Maps property id to list of :py:class:`.WikidataSnak`
    """

    def __init__(self, reference_dict: typedefs.ReferenceDict) -> None:
        self._validate_reference_dict(reference_dict)
        self._reference_dict = reference_dict

        self.referencehash = reference_dict["hash"]
        self.snaks = OrderedDict()  # type: OrderedDict
        for property_id in reference_dict["snaks-order"]:
            self.snaks[property_id] = [
                WikidataSnak(snak_dict) for snak_dict in reference_dict["snaks"][property_id]
            ]

    def _validate_reference_dict(self, reference_dict: typedefs.ReferenceDict) -> None:
        """Raise excpetions if reference_dict is not valid."""
        _REQUIRED_KEYS = ["hash", "snaks", "snaks-order"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in reference_dict:
                raise ValueError(
                    "required reference_dict keys are {} but only found {}".format(
                        _REQUIRED_KEYS, list(reference_dict.keys())
                    )
                )

    def __str__(self) -> str:
        return "WikidataReference(hash={}, snaks={})".format(self.referencehash, self.snaks)

    def __repr__(self) -> str:
        return self.__str__()


class WikidataQualifier:
    """A qualifier about a claim about a Wikidata Entity.

    See: https://www.wikidata.org/wiki/Help:Qualifiers

    This class can be initialized from an entity dictionary as,

    .. code-block:: python

      >>> qualifier_dict = q42_dict['claims']['P69'][0]['qualifiers']['P582'][0]
      >>> wikidata_qualifier = WikidataQualifier(qualifier_dict)


    Parameters
    ----------
    qualifier_dict
      A dictionary representing a Wikidata qualifier.
      See `the wikibase JSON data model docs`_ for a description
      of the format.


    Attributes
    ----------
    qualifierhash: str
      Unique id for this qualifier
    snak: WikidataSnak
      The snak for this qualifier

    """

    def __init__(self, qualifier_dict: typedefs.QualifierDict) -> None:
        self._validate_qualifier_dict(qualifier_dict)
        self._qualifier_dict = qualifier_dict

        self.qualifierhash = qualifier_dict["hash"]
        self.snak = WikidataSnak(qualifier_dict)

    def _validate_qualifier_dict(self, qualifier_dict: typedefs.QualifierDict) -> None:
        """Raise excpetions if qualifier_dict is not valid."""
        _REQUIRED_KEYS = ["hash", "snaktype", "property", "datatype"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in qualifier_dict:
                raise ValueError(
                    "required qualifier_dict keys are {} but only found {}".format(
                        _REQUIRED_KEYS, list(qualifier_dict.keys())
                    )
                )

    def __str__(self) -> str:
        return "WikidataQualifier(hash={}, snak={})".format(self.qualifierhash, self.snak)

    def __repr__(self) -> str:
        return self.__str__()


class WikidataClaim:
    """A claim aka statement about a Wikidata Entity.

    From the Wikibase data model docs,

        "Statements describe the claim of a statement and list references for this claim.
        Every Statement refers to one particular Entity, called the subject of the Statement.
        There is always one main Snak that forms the most important part of the statement.
        Moreover, there can be zero or more additional PropertySnaks that describe the Statement
        in more detail. These qualifier Snaks (or "qualifiers" for short) store additional
        information that does not directly refer to the subject (e.g., the time at which the
        main part of the statement was valid). References are provided as a list (the order is
        significant in some contexts, especially for displaying a main reference)."

        -- https://www.mediawiki.org/wiki/Wikibase/DataModel


    This class can be initialized from an entity dictionary as,

    .. code-block:: python

      >>> claim_dict = q42_dict['claims']['P551'][0]
      >>> wikidata_claim = WikidataClaim(claim_dict)


    Parameters
    ----------
    claim_dict: dict
      A dictionary representing a Wikidata claim.
      See `the wikibase JSON data model docs`_ for a description
      of the format.


    Attributes
    ----------
    claim_id: str
      Unique id for this claim
    property_id: PropertyId
      A Wikiata property id (e.g. "P551")
    claim_type: str
      One of ["claim", "statement"] (statements may have references)
    rank: str
      One of ["preferred", "normal", "deprecated"]
    mainsnak: :py:class:`.WikidataSnak`
      The mainsnak of this claim
    qualifiers: collections.OrderedDict
      Maps property id to list of :py:class:`WikidataQualifier`
    references: list
      A list of :py:class:`WikidataReference`
    qualifiers_order: list
      The order of the property ids in qualifiers


    .. _the wikibase JSON data model docs: https://www.mediawiki.org/wiki/Wikibase/DataModel/JSON
    """

    def __init__(self, claim_dict: typedefs.ClaimDict) -> None:
        self._validate_claim_dict(claim_dict)
        self._claim_dict = claim_dict
        self.property_id = self.mainsnak.property_id

        self.qualifiers = (
            OrderedDict()
        )  # type: OrderedDict[typedefs.PropertyId, List[WikidataQualifier]]
        self.qualifiers_order = claim_dict.get("qualifiers-order", [])
        if "qualifiers" in claim_dict:
            for property_id in self.qualifiers_order:
                qualifier_dicts = claim_dict["qualifiers"][property_id]
                self.qualifiers[property_id] = [WikidataQualifier(qd) for qd in qualifier_dicts]

        self.references = []  # type: List[WikidataReference]
        if "references" in claim_dict:
            for reference_dict in claim_dict["references"]:
                self.references.append(WikidataReference(reference_dict))

    def _validate_claim_dict(self, claim_dict: typedefs.ClaimDict) -> None:
        """Raise excpetions if claim_dict is not valid."""
        _REQUIRED_KEYS = ["id", "type", "rank", "mainsnak"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in claim_dict:
                raise ValueError(
                    "required claim_dict keys are {} but only found {}".format(
                        _REQUIRED_KEYS, list(claim_dict.keys())
                    )
                )
        self.claim_id = claim_dict["id"]
        self.claim_type = claim_dict["type"]
        self.rank = claim_dict["rank"]
        self.mainsnak = WikidataSnak(claim_dict["mainsnak"])

    def __str__(self) -> str:
        return "WikidataClaim(type={}, rank={}, mainsnak={}, qualifiers={})".format(
            self.claim_type, self.rank, self.mainsnak, self.qualifiers
        )

    def __repr__(self) -> str:
        return self.__str__()


class WikidataClaimGroup(Sequence):
    """A sequence of :py:class:`WikidataClaim` instances with a common property id.

    For example the claim group for "Douglas Adams" (Q42) with property "residence" (P551)
    has three elements.  This class can be initialized from an entity dictionary as,

    .. code-block:: python

      >>> claim_group = WikidataClaimGroup(q42_dict['claims']['P551'])


    Parameters
    ----------
    claim_list: list
      A list of claim dictionaries representing a Wikidata claim group.
      See `the wikibase JSON data model docs`_ for a description
      of the format.


    .. _the wikibase JSON data model docs: https://www.mediawiki.org/wiki/Wikibase/DataModel/JSON
    """

    def __init__(self, claim_list: typedefs.ClaimList) -> None:
        super(WikidataClaimGroup, self).__init__()
        self._validate_claim_list(claim_list)
        self._claim_list = claim_list
        self._claims = [WikidataClaim(claim_dict) for claim_dict in claim_list]

        property_ids = set([claim.mainsnak.property_id for claim in self._claims])
        self.property_id = None  # type: Union[typedefs.PropertyId, None]
        if len(property_ids) == 1:
            self.property_id = property_ids.pop()
        elif len(property_ids) == 0:
            self.property_id = None
        else:
            raise ValueError(
                "claims in a claim list must all have the same property id but found multiple property ids {}".format(
                    property_ids
                )
            )

    def _validate_claim_list(self, claim_list: typedefs.ClaimList) -> None:
        """Raise excpetions if claim_list is not valid."""
        if not isinstance(claim_list, list):
            raise TypeError("claim_list must be a list but got {}.".format(type(claim_list)))

    @overload
    def __getitem__(self, indx: int) -> WikidataClaim:
        ...

    @overload
    def __getitem__(self, indx: slice) -> List[WikidataClaim]:
        ...

    def __getitem__(self, indx: Union[int, slice]) -> Union[WikidataClaim, List[WikidataClaim]]:
        return self._claims[indx]

    def __len__(self) -> int:
        return len(self._claims)

    def __str__(self) -> str:
        return "WikidataClaimGroup(property_id={}, claims={})".format(
            self.property_id, self._claims
        )

    def __repr__(self) -> str:
        return self.__str__()



# Copyright 2019 Kensho Technologies, LLC.
"""Module for Wikidata Entities."""

from typing import Dict, List, Union

from qwikidata import typedefs
from qwikidata.claim import WikidataClaimGroup


class EntityMixin:
    """Mixin for all entities.

    .. seealso::

        * :py:class:`WikidataItem`
        * :py:class:`WikidataProperty`
        * :py:class:`WikidataLexeme`

    """

    @staticmethod
    def _validate_entity_dict(entity_dict: typedefs.EntityDict) -> None:
        """Raise excpetions if entity_dict is not valid."""
        _REQUIRED_KEYS = ["id", "type"]
        if not isinstance(entity_dict, dict):
            raise TypeError(
                "entity_dict must be a dictionary but got {}.".format(type(entity_dict))
            )
        for req_key in _REQUIRED_KEYS:
            if req_key not in entity_dict:
                raise ValueError(
                    "required entity_dict keys are {} but only found {}.".format(
                        _REQUIRED_KEYS, list(entity_dict.keys())
                    )
                )


class LabelDescriptionAliasMixin:
    """Mixin for entities with labels, descriptions, and aliases.

    .. seealso::

        * :py:class:`WikidataItem`
        * :py:class:`WikidataProperty`

    """

    _entity_dict = None  # type: typedefs.EntityDict

    @staticmethod
    def _validate_label_desc_alias_dict(
        label_desc_alias_dict: Union[typedefs.ItemDict, typedefs.PropertyDict]
    ) -> None:
        """Raise excpetions if label_desc_alias_dict is not valid."""
        _REQUIRED_KEYS = ["labels", "descriptions", "aliases"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in label_desc_alias_dict:
                raise ValueError(
                    "required label_desc_alias_dict keys are {} but only found {}.".format(
                        _REQUIRED_KEYS, list(label_desc_alias_dict.keys())
                    )
                )

    def get_label(self, lang: typedefs.LanguageCode = typedefs.LanguageCode("en")) -> str:
        """Get label (primary name for this entity) in a specific language.

        See: https://www.wikidata.org/wiki/Help:Label

        Parameters
        ----------
        lang
          Find the label in this language.
        """
        if isinstance(self._entity_dict["labels"], dict) and lang in self._entity_dict["labels"]:
            return self._entity_dict["labels"][lang]["value"]
        else:
            return ""

    def get_description(self, lang: typedefs.LanguageCode = typedefs.LanguageCode("en")) -> str:
        """Get a brief description of this entity in a specific language.

        See: https://www.wikidata.org/wiki/Help:Description

        Parameters
        ----------
        lang
          Find the description in this language.
        """
        if (
            isinstance(self._entity_dict["descriptions"], dict)
            and lang in self._entity_dict["descriptions"]
        ):
            return self._entity_dict["descriptions"][lang]["value"]
        else:
            return ""

    def get_aliases(self, lang: typedefs.LanguageCode = typedefs.LanguageCode("en")) -> List[str]:
        """Get alternative names for this entity in a specific language.

        See: https://www.wikidata.org/wiki/Help:Aliases

        Parameters
        ----------
        lang
          Find aliases in this language.
        """
        if isinstance(self._entity_dict["aliases"], dict) and lang in self._entity_dict["aliases"]:
            return [el["value"] for el in self._entity_dict["aliases"][lang]]
        else:
            return []


class ClaimsMixin:
    """Mixin for entities with top level claims (aka statements).

    See: https://www.wikidata.org/wiki/Help:Statements

    .. seealso::

        * :py:class:`WikidataItem`
        * :py:class:`WikidataProperty`
        * :py:class:`WikidataLexeme`
        * :py:class:`WikidataForm`
        * :py:class:`WikidataSense`
    """

    _entity_dict = None  # type: typedefs.EntityDict

    @staticmethod
    def _validate_claim_dict(claim_dict: typedefs.EntityDict) -> None:
        """Raise excpetions if claim_dict is not valid."""
        _REQUIRED_KEYS = ["claims"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in claim_dict:
                raise ValueError(
                    "required claim_dict keys are {}. but only found {}".format(
                        _REQUIRED_KEYS, list(claim_dict.keys())
                    )
                )

    def get_claim_groups(self) -> Dict[typedefs.PropertyId, WikidataClaimGroup]:
        """Get all claim groups about this entity."""
        if isinstance(self._entity_dict["claims"], dict):
            claims = {
                property_id: WikidataClaimGroup(claim_list)
                for property_id, claim_list in self._entity_dict["claims"].items()
            }
            return claims
        else:
            return {}

    def get_claim_group(self, property_id: typedefs.PropertyId) -> WikidataClaimGroup:
        """Get the claim group corresponding to a given property id.

        Parameters
        ----------
        property_id
            the string representing the property ID of the claim group to return
        """
        if not isinstance(self._entity_dict["claims"], dict):
            return WikidataClaimGroup([])

        claim_list = self._entity_dict["claims"].get(property_id, None)
        if claim_list is None:
            return WikidataClaimGroup([])
        else:
            return WikidataClaimGroup(claim_list)

    def get_truthy_claim_groups(self) -> Dict[typedefs.PropertyId, WikidataClaimGroup]:
        """Get all truthy claim groups about this entity.

        Truthy is defined in the Wikidata RDF dump format docs,

            "Truthy statements represent statements that have the best non-deprecated rank for a
            given property. Namely, if there is a preferred statement for a property P, then only
            preferred statements for P will be considered truthy. Otherwise, all normal-rank
            statements for P are considered truthy."

            -- `RDF dump format docs on truthy statements`_


        .. _RDF dump format docs on truthy statements: https://www.mediawiki.org/wiki/Wikibase/Indexing/RDF_Dump_Format#Truthy_statements
        """
        if isinstance(self._entity_dict["claims"], dict):
            claims = {}
            property_ids = self._entity_dict["claims"].keys()
            for property_id in property_ids:
                claims[property_id] = self.get_truthy_claim_group(property_id)
            return claims
        else:
            return {}

    def get_truthy_claim_group(self, property_id: typedefs.PropertyId) -> WikidataClaimGroup:
        """Get truthy claims from the claim group corresponding to a given property id.

        Truthy is defined in the Wikidata RDF dump format docs,

            "Truthy statements represent statements that have the best non-deprecated rank for a
            given property. Namely, if there is a preferred statement for a property P, then only
            preferred statements for P will be considered truthy. Otherwise, all normal-rank
            statements for P are considered truthy."

            -- `RDF dump format docs on truthy statements`_


        Parameters
        ----------
        property_id
            the string representing the property ID of the claim group to return


        .. _RDF dump format docs on truthy statements: https://www.mediawiki.org/wiki/Wikibase/Indexing/RDF_Dump_Format#Truthy_statements
        """
        claim_group = self.get_claim_group(property_id)

        truthy_claim_dicts = [
            claim._claim_dict for claim in claim_group if claim.rank.lower() == "preferred"
        ]

        if len(truthy_claim_dicts) == 0:
            truthy_claim_dicts = [
                claim._claim_dict for claim in claim_group if claim.rank.lower() != "deprecated"
            ]

        return WikidataClaimGroup(truthy_claim_dicts)


class WikidataItem(LabelDescriptionAliasMixin, ClaimsMixin, EntityMixin):
    """Class for Wikidata Items.

    Parameters
    ----------
    item_dict
      A dictionary representation of a Wikidata Item.
      See `the wikibase JSON data model docs`_ for a description
      of the dictionary format.


    .. seealso::

      Wikidata docs on items,

        * https://www.wikidata.org/wiki/Help:Items


      Ways to generate item dictionaries within qwikidata.

        * :py:class:`qwikidata.json_dump.WikidataJsonDump`
        * :py:func:`qwikidata.linked_data_interface.get_entity_dict_from_api`


    .. _the wikibase JSON data model docs: https://www.mediawiki.org/wiki/Wikibase/DataModel/JSON
    """

    def __init__(self, item_dict: typedefs.ItemDict) -> None:
        self._validate_item_dict(item_dict)
        self._entity_dict = item_dict
        self.entity_id = item_dict["id"]
        self.entity_type = item_dict["type"]

    def _validate_item_dict(self, item_dict: typedefs.ItemDict) -> None:
        """Raise excpetions if item_dict is not valid."""
        self._validate_entity_dict(item_dict)
        if item_dict["type"] != "item":
            raise ValueError(
                "item_dict['type'] must be 'item' but found '{}'".format(item_dict["type"])
            )
        self._validate_label_desc_alias_dict(item_dict)
        self._validate_claim_dict(item_dict)

    def get_sitelinks(self, prefix: str = "enwiki") -> Dict[str, typedefs.SitelinkDict]:
        """Get Wikimedia sitelinks for this item.

        Parameters
        ----------
        prefix
          filters to sitelinks that begin with `prefix`.  can filter by language (e.g. "en")
          or by language and site (e.g. "enwiki").

        Returns
        -------
        dict
          A dictionary with site names as keys and sitelink dictionaries as values.
        """
        if isinstance(self._entity_dict["sitelinks"], dict):
            return {k: v for k, v in self._entity_dict["sitelinks"].items() if k.startswith(prefix)}
        else:
            return {}

    def get_enwiki_title(self) -> str:
        """Get english language wikipedia page title."""
        if (
            isinstance(self._entity_dict["sitelinks"], dict)
            and "enwiki" in self._entity_dict["sitelinks"]
        ):
            return self._entity_dict["sitelinks"]["enwiki"]["title"]
        else:
            return ""

    def __str__(self) -> str:
        return "WikidataItem(label={}, id={}, description={}, aliases={}, enwiki_title={})".format(
            self.get_label(),
            self.entity_id,
            self.get_description(),
            self.get_aliases(),
            self.get_enwiki_title(),
        )

    def __repr__(self) -> str:
        return self.__str__()


class WikidataProperty(LabelDescriptionAliasMixin, ClaimsMixin, EntityMixin):
    """Class for Wikidata Properties.

    Parameters
    ----------
    property_dict
      A dictionary representation of a Wikidata Property.
      See `the wikibase JSON data model docs`_ for a description
      of the dictionary format.


    .. seealso::

      Wikidata docs on properties,

        * https://www.wikidata.org/wiki/Help:Properties


      Ways to generate property dictionaries within kwikimedia.

        * :py:class:`qwikidata.json_dump.WikidataJsonDump`
        * :py:func:`qwikidata.linked_data_interface.get_entity_dict_from_api`


    .. _the wikibase JSON data model docs: https://www.mediawiki.org/wiki/Wikibase/DataModel/JSON
    """

    def __init__(self, property_dict: typedefs.PropertyDict) -> None:
        self._validate_property_dict(property_dict)
        self._entity_dict = property_dict
        self.entity_id = property_dict["id"]
        self.entity_type = property_dict["type"]

    def _validate_property_dict(self, property_dict: typedefs.PropertyDict) -> None:
        """Raise excpetions if property_dict is not valid."""
        self._validate_entity_dict(property_dict)
        if property_dict["type"] != "property":
            raise ValueError(
                "property_dict['type'] must be 'property' but found '{}'".format(
                    property_dict["type"]
                )
            )
        self._validate_label_desc_alias_dict(property_dict)
        self._validate_claim_dict(property_dict)

    def __str__(self) -> str:
        return "WikidataProperty(label={}, id={}, description={}, aliases={})".format(
            self.get_label(), self.entity_id, self.get_description(), self.get_aliases()
        )

    def __repr__(self) -> str:
        return self.__str__()


class WikidataForm(ClaimsMixin):
    """A form associated with a Wikidata Lexeme.

    See: https://www.mediawiki.org/wiki/Extension:WikibaseLexeme/Data_Model#Form

    This class can be initialized from a lexeme dictionary as,

    .. code-block:: python

      >>> form_dict = l3354_dict['forms'][0]
      >>> wikidata_form = WikidataForm(form_dict)


    Parameters
    ----------
    form_dict
      A dictionary representing a Wikidata Lexeme form.


    Attributes
    ----------
    form_id: str
      Unique id for this form (e.g. 'L3354-F1')
    grammatical_features: list
      List of item ids representing grammatical categories (e.g. present tense, first person, ...)
    """

    def __init__(self, form_dict: typedefs.FormDict) -> None:
        self._validate_form_dict(form_dict)
        self._form_dict = form_dict

        self.form_id = form_dict["id"]
        self.grammatical_features = form_dict["grammaticalFeatures"]

    def _validate_form_dict(self, form_dict: typedefs.FormDict) -> None:
        """Raise excpetions if form_dict is not valid."""
        _REQUIRED_KEYS = ["id", "representations", "grammaticalFeatures", "claims"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in form_dict:
                raise ValueError(
                    "required form_dict keys are {} but only found {}".format(
                        _REQUIRED_KEYS, list(form_dict.keys())
                    )
                )

    def get_representation(self, lang: typedefs.LanguageCode = typedefs.LanguageCode("en")) -> str:
        """Get representation of this form in a given language.

        See: https://www.mediawiki.org/wiki/Extension:WikibaseLexeme/Data_Model#Representation

        Parameters
        ----------
        lang
          Find the representation in this language.
        """
        if (
            isinstance(self._form_dict["representations"], dict)
            and lang in self._form_dict["representations"]
        ):
            return self._form_dict["representations"][lang]["value"]
        else:
            return ""

    def __str__(self) -> str:
        return "WikidataForm(form_id={}, representation={}, grammatical_features={})".format(
            self.form_id, self.get_representation(), self.grammatical_features
        )

    def __repr__(self) -> str:
        return self.__str__()


class WikidataSense(ClaimsMixin):
    """A sense associated with a Wikidata Lexeme.

    See: https://www.mediawiki.org/wiki/Extension:WikibaseLexeme/Data_Model#Sense

    This class can be initialized from a lexeme dictionary as,

    .. code-block:: python

      >>> sense_dict = l3354_dict['senses'][0]
      >>> wikidata_sense = WikidataSense(sense_dict)


    Parameters
    ----------
    sense_dict
      A dictionary representing a Wikidata Lexeme sense.


    Attributes
    ----------
    sense_id: str
      Unique id for this sense (e.g. 'L3354-S1')
    """

    def __init__(self, sense_dict: typedefs.SenseDict) -> None:
        self._validate_sense_dict(sense_dict)
        self._sense_dict = sense_dict

        self.sense_id = sense_dict["id"]

    def _validate_sense_dict(self, sense_dict: typedefs.SenseDict) -> None:
        """Raise excpetions if sense_dict is not valid."""
        _REQUIRED_KEYS = ["id", "glosses", "claims"]
        for req_key in _REQUIRED_KEYS:
            if req_key not in sense_dict:
                raise ValueError(
                    "required sense_dict keys are {} but only found {}".format(
                        _REQUIRED_KEYS, list(sense_dict.keys())
                    )
                )

    def get_gloss(self, lang: typedefs.LanguageCode = typedefs.LanguageCode("en")) -> str:
        """Get gloss of this sense in a given language.

        See: https://www.mediawiki.org/wiki/Extension:WikibaseLexeme/Data_Model#Gloss

        Parameters
        ----------
        lang
          Find the gloss in this language.
        """
        if isinstance(self._sense_dict["glosses"], dict) and lang in self._sense_dict["glosses"]:
            return self._sense_dict["glosses"][lang]["value"]
        else:
            return ""

    def __str__(self) -> str:
        return "WikidataSense(sense_id={}, gloss={})".format(self.sense_id, self.get_gloss())

    def __repr__(self) -> str:
        return self.__str__()


class WikidataLexeme(ClaimsMixin, EntityMixin):
    """Class for Wikidata Lexeme.

    Parameters
    ----------
    lexeme_dict
      A dictionary representation of a Wikidata Lexeme.
      See `the wikibase Lexeme JSON data model docs`_ for a description
      of the dictionary format.


    .. seealso::

      Wikidata docs on lexemes,

        * https://www.wikidata.org/wiki/Wikidata:Lexicographical_data


      Ways to generate lexeme dictionaries within qwikidata.

        * :py:class:`qwikidata.json_dump.WikidataJsonDump`
        * :py:func:`qwikidata.linked_data_interface.get_entity_dict_from_api`


    .. _the wikibase Lexeme JSON data model docs: https://www.mediawiki.org/wiki/Extension:WikibaseLexeme/Data_Model
    """

    def __init__(self, lexeme_dict: typedefs.LexemeDict) -> None:
        self._validate_lexeme_dict(lexeme_dict)
        self._entity_dict = lexeme_dict  # type: typedefs.LexemeDict
        self.entity_id = lexeme_dict["id"]
        self.entity_type = lexeme_dict["type"]
        self.language = lexeme_dict["language"]
        self.lexical_category = lexeme_dict["lexicalCategory"]

    def _validate_lexeme_dict(self, lexeme_dict: typedefs.LexemeDict) -> None:
        """Raise excpetions if lexeme_dict is not valid."""
        self._validate_entity_dict(lexeme_dict)
        if lexeme_dict["type"] != "lexeme":
            raise ValueError(
                "lexeme_dict['type'] must be 'lexeme' but found '{}'".format(lexeme_dict["type"])
            )
        self._validate_claim_dict(lexeme_dict)

        _REQUIRED_KEYS = ["lemmas", "lexicalCategory", "language", "forms", "senses"]

        for req_key in _REQUIRED_KEYS:
            if req_key not in lexeme_dict:
                raise ValueError(
                    "required lexeme_dict keys are  {} but only found {}".format(
                        _REQUIRED_KEYS, list(lexeme_dict.keys())
                    )
                )

    def get_lemma(self, lang: typedefs.LanguageCode = typedefs.LanguageCode("en")) -> str:
        """Get lemma (primary name for this lexeme) in a specific language.

        See: https://www.mediawiki.org/wiki/Extension:WikibaseLexeme/Data_Model#Lemma

        Parameters
        ----------
        lang
          Find the lemma in this language.
        """
        if isinstance(self._entity_dict["lemmas"], dict) and lang in self._entity_dict["lemmas"]:
            return self._entity_dict["lemmas"][lang]["value"]
        else:
            return ""

    def get_forms(self) -> List[WikidataForm]:
        """Get the set of forms assocaited with this Lexeme."""
        return [WikidataForm(form_dict) for form_dict in self._entity_dict["forms"]]

    def get_senses(self) -> List[WikidataSense]:
        """Get the set of senses assocaited with this Lexeme."""
        return [WikidataSense(sense_dict) for sense_dict in self._entity_dict["senses"]]

    def __str__(self) -> str:
        return "WikidataLexeme(lemma={}, id={}, language={}, lexical_category={}, forms={}, senses={})".format(
            self.get_lemma(),
            self.entity_id,
            self.language,
            self.lexical_category,
            self.get_forms(),
            self.get_senses(),
        )

    def __repr__(self) -> str:
        return self.__str__()


WikidataEntity = Union[WikidataItem, WikidataProperty, WikidataLexeme]



# Copyright 2019 Kensho Technologies, LLC.
"""Module for Wikidata Datavalues."""

import re
from typing import Dict, Union

from qwikidata import typedefs


def _validate_datavalue_dict(datavalue_dict: typedefs.DatavalueDict) -> None:
    """Raise excpetions if datavalue_dict is not valid."""
    _REQUIRED_KEYS = ["type", "value"]
    for req_key in _REQUIRED_KEYS:
        if req_key not in datavalue_dict:
            raise ValueError(
                "required datavalue_dict keys are {} but only found {}".format(
                    _REQUIRED_KEYS, list(datavalue_dict.keys())
                )
            )

    _VALID_TYPES = frozenset(
        [
            "globecoordinate",
            "monolingualtext",
            "quantity",
            "string",
            "time",
            "wikibase-entityid",
            "wikibase-unmapped-entityid",
        ]
    )
    if datavalue_dict["type"] not in _VALID_TYPES:
        raise ValueError(
            "datavalue datatype={} not in valid datatypes {}.".format(
                datavalue_dict["type"], _VALID_TYPES
            )
        )


class GlobeCoordinate:
    """Class for `globecoordinate` datavalues.

    In this class, the `value` attribute is a mapping with the following keys,

    * **latitude** (`float` or `str`): the latitude part of the coordinate in degrees
    * **longitude** (`float` or `str`): the longitude part of the coordinate in degrees
    * **precision** (`float` or `str`): the coordinate's precision, in (fractions of)
      degrees
    * **globe** (`str`): the URI of a reference globe. This would typically refer to a
      data item on wikidata.org. This is usually just an indication of the celestial body
      (e.g. Q2 = earth), but could be more specific, like WGS 84 or ED50.
    * **altitude** (`float` or `str`): Deprecated and no longer used. Will be dropped in the future.


    Attributes
    ----------
    value: dict
      mapping that represents a location on a globe
    datatype: str
      `globecoordinate`
    """

    def __init__(self, datavalue_dict: typedefs.GlobeCoordinateDatavalueDict) -> None:
        _validate_datavalue_dict(datavalue_dict)
        self._datavalue_dict = datavalue_dict
        self.datatype = datavalue_dict["type"]
        self.value = datavalue_dict["value"]

    def __str__(self) -> str:
        return "GlobeCoordinate(latitude={}, longitude={})".format(
            self.value["latitude"], self.value["longitude"]
        )


class MonolingualText:
    """Class for `monolingualtext` datavalues.

    In this class, the `value` attribute is a mapping with the following keys,

    * **text** (`str`): a string literal
    * **language** (`str`): a `Wikidata language code`_


    .. _Wikidata language code: https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all


    Attributes
    ----------
    value: dict
      mapping that represents text in a specific language
    datatype: str
      `monolingualtext`
    """

    def __init__(self, datavalue_dict: typedefs.MonolingualTextDatavalueDict) -> None:
        _validate_datavalue_dict(datavalue_dict)
        self._datavalue_dict = datavalue_dict
        self.datatype = datavalue_dict["type"]
        self.value = datavalue_dict["value"]

    def __str__(self) -> str:
        return "MongolingualText(text={}, language={})".format(
            self.value["text"], self.value["language"]
        )


class Quantity:
    """Class for `quantity` datavalues.

    In this class, the `value` attribute is a mapping with the following keys,

    * **amount** (`str`): The nominal value of the quantity, as an arbitrary precision decimal
      string. The string always starts with a character indicating the sign of the value, either
      "+" or "-".
    * **upperBound** (`str`): Optionally, the upper bound of the quantity's uncertainty interval,
      using the same notation as the amount field. If not given or null, the uncertainty (or
      precision) of the quantity is not known. If the upperBound field is given, the lowerBound
      field must also be given.
    * **lowerBound** (`str`): Optionally, the lower bound of the quantity's uncertainty interval,
      using the same notation as the amount field. If not given or null, the uncertainty (or
      precision) of the quantity is not known. If the lowerBound field is given, the upperBound
      field must also be given.
    * **unit** (`str`): the URI of a unit (or "1" to indicate a unit-less quantity). This would
      typically refer to a data item on wikidata.org, e.g. http://www.wikidata.org/entity/Q712226
      for "square kilometer".


    Attributes
    ----------
    value: dict
      mapping that represents a numeric quantity
    datatype: str
      `quantity`
    """

    def __init__(self, datavalue_dict: typedefs.QuantityDatavalueDict) -> None:
        _validate_datavalue_dict(datavalue_dict)
        self._datavalue_dict = datavalue_dict
        self.datatype = datavalue_dict["type"]
        self.value = datavalue_dict["value"]

    def __str__(self) -> str:
        return "Quantity(amount={}, unit={})".format(self.value["amount"], self.value["unit"])


class String:
    """Class for `string` datavalues.

    Attributes
    ----------
    value: str
      a string literal
    datatype: str
      `string`
    """

    def __init__(self, datavalue_dict: typedefs.StringDatavalueDict) -> None:
        _validate_datavalue_dict(datavalue_dict)
        self._datavalue_dict = datavalue_dict
        self.datatype = datavalue_dict["type"]
        self.value = datavalue_dict["value"]

    def __str__(self) -> str:
        return "String(value={})".format(self.value)


class Time:
    """Class for `time` datavalues.

    In this class, the `value` attribute is a mapping with the following keys,

    * **time** (`str`): the format and interpretation of this string depends on the calendar model.
      Currently, only Julian and Gregorian dates are supported.  The format used for Gregorian
      and Julian dates use a notation resembling ISO 8601. E.g. "+1994-01-01T00:00:00Z". The
      year is represented by at least four digits, zeros are added on the left side as needed.
      Years BCE are represented as negative numbers, using the historical numbering, in which
      year 0 is undefined, and the year 1 BCE is represented as -0001, the year 44 BCE is
      represented as -0044, etc., like XSD 1.0 (ISO 8601:1988) does.
      Month and day may be 00 if they are unknown or insignificant. The day of the month
      may have values between 0 and 31 for any month, to accommodate "leap dates" like February
      30. Hour, minute, and second are currently unused and should always be 00.

    * **timezone** (`int`): Signed integer. Currently unused, and should always be 0.

    * **calendarmodel** (`str`): URI of a calendar model, such as gregorian or julian. Typically
      given as the URI of a data item on the repository

    * **precision** (`int`): To what unit is the given date/time significant? Given as an integer
      indicating one of the following units:

        * 0: 1 Gigayear
        * 1: 100 Megayears
        * 2: 10 Megayears
        * 3: Megayear
        * 4: 100 Kiloyears
        * 5: 10 Kiloyears
        * 6: Kiloyear
        * 7: 100 years
        * 8: 10 years
        * 9: years
        * 10: months
        * 11: days
        * 12: hours (unused)
        * 13: minutes (unused)
        * 14: seconds (unused)

      Note that the precision should be read as an indicator of the significant parts of the date
      string, it does not directly specify an interval. That is, 1988-07-13T00:00:00 with
      precision 8 (decade) will be interpreted as 198?-??-?? and rendered as "1980s".
      1981-01-21T00:00:00 with precision 8 would have the exact same interpretation. Thus the two
      dates are equivalent, since year, month, and days are treated as insignificant.

    * **before** (`int`): Beginning of an uncertainty range, given in the unit defined by the
      precision field. This cannot be used to represent a duration. (Currently unused, may be
      dropped in the future)

    * **after** (`int`): End of an uncertainty range, given in the unit defined by the precision
      field. This cannot be used to represent a duration. (Currently unused, may be dropped in the
      future)


    Attributes
    ----------
    value: dict
      mapping that represents a time
    datatype: str
      `time`
    """

    def __init__(self, datavalue_dict: typedefs.TimeDatavalueDict) -> None:
        _validate_datavalue_dict(datavalue_dict)
        self._datavalue_dict = datavalue_dict
        self.datatype = datavalue_dict["type"]
        self.value = datavalue_dict["value"]
        self.STANDARD_DATE_REGEX = re.compile(
            r"""
            (?P<year>[+-]?\d+?)-
            (?P<month>\d\d)-
            (?P<day>\d\d)T
            (?P<hour>\d\d):
            (?P<minute>\d\d):
            (?P<second>\d\d)Z?""",
            re.VERBOSE,
        )

    def __str__(self) -> str:
        return "Time(time={}, precision={})".format(self.value["time"], self.value["precision"])

    def get_parsed_datetime_dict(self) -> Dict[str, int]:
        """Return a dictionary representation of the datavalue.

        Given a Wikidata time string, extract the year, month, and day.

        Time strings look like this, for examples: '+1838-01-01T00:00:00Z'.

        See: https://www.wikidata.org/wiki/Help:Dates

        TODO: Handle truncated dates (like 20 for 20th century)
        TODO: Allow for partial match if we only know some information like the year

        Returns
        -------
        dict
            a dictionary representing the timestring's year, month, and date
        """
        datetime_dict = {}  # type: Dict[str, int]
        timestring = self.value["time"]
        match = self.STANDARD_DATE_REGEX.fullmatch(timestring)
        if match:

            datetime_dict = {
                "year": int(match.group("year")),
                "month": int(match.group("month")),
                "day": int(match.group("day")),
                "hour": int(match.group("hour")),
                "minute": int(match.group("minute")),
                "second": int(match.group("second")),
            }

        return datetime_dict


class WikibaseEntityId:
    """Class for `wikibase-entityid` datavalues.

    In this class, the `value` attribute is a mapping with the following keys,

      * **entity-type** (`str`): one of ["item", "property"]
      * **id** (`str`): string form of entity id (e.g. "Q42")
      * **numeric-id** (`int`): integer form of entity id (e.g. 42)

    Attributes
    ----------
    value: dict
      mapping that references a Wikidata entity
    datatype: str
      `wikibase-entityid`
    """

    def __init__(self, datavalue_dict: typedefs.WikibaseEntityIdDatavalueDict) -> None:
        _validate_datavalue_dict(datavalue_dict)
        self._datavalue_dict = datavalue_dict
        self.datatype = datavalue_dict["type"]
        self.value = datavalue_dict["value"]

    def __str__(self) -> str:
        return "WikibaseEntityId(id={})".format(self.value["id"])


class WikibaseUnmappedEntityId:
    """Class for `wikibase-unmapped-entityid` datavalues.

    In this class, the `value` attribute is a string representing an
    unmapped wikibase entity id.

    Attributes
    ----------
    value: str
      string that references an unmapped entity id
    datatype: str
      `wikibase-unmapped-entityid`
    """

    def __init__(self, datavalue_dict: typedefs.WikibaseUnmappedEntityIdDatavalueDict) -> None:
        _validate_datavalue_dict(datavalue_dict)
        self._datavalue_dict = datavalue_dict
        self.datatype = datavalue_dict["type"]
        self.value = datavalue_dict["value"]

    def __str__(self) -> str:
        return "WikibaseUnmappedEntityId(value={})".format(self.value)


_DATAVALUE_TYPE_TO_CLASS = {
    "globecoordinate": GlobeCoordinate,
    "monolingualtext": MonolingualText,
    "quantity": Quantity,
    "string": String,
    "time": Time,
    "wikibase-entityid": WikibaseEntityId,
    "wikibase-unmapped-entityid": WikibaseUnmappedEntityId,
}


WikidataDatavalue = Union[
    GlobeCoordinate,
    MonolingualText,
    Quantity,
    String,
    Time,
    WikibaseEntityId,
    WikibaseUnmappedEntityId,
]


def get_datavalue_from_snak_dict(snak_dict: typedefs.SnakDict) -> Union[WikidataDatavalue, None]:
    """Return a Wikidata Datavalue from a snak dictionary."""
    if snak_dict["snaktype"] == "value":
        datavalue_class = _DATAVALUE_TYPE_TO_CLASS[str(snak_dict["datavalue"]["type"])]
        return datavalue_class(snak_dict["datavalue"])
    elif snak_dict["snaktype"] == "somevalue" or snak_dict["snaktype"] == "novalue":
        return None
    else:
        raise ValueError(
            'snaktype must be one of ["value", "somevalue", "novalue"] but got {}.'.format(
                snak_dict["snaktype"]
            )
        )
